package ctbrec.ui.sites.chaturbate;

import ctbrec.Config;
import ctbrec.Model;
import ctbrec.io.HttpException;
import ctbrec.sites.chaturbate.Chaturbate;
import ctbrec.ui.SiteUiFactory;
import ctbrec.ui.tabs.PaginatedScheduledService;
import javafx.concurrent.Task;
import okhttp3.Request;
import okhttp3.Response;
import org.json.JSONArray;
import org.json.JSONObject;
import org.jsoup.Jsoup;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Locale;

import static ctbrec.io.HttpConstants.*;

public class ChaturbateUpdateService extends PaginatedScheduledService {

    private static final Logger LOG = LoggerFactory.getLogger(ChaturbateUpdateService.class);
    private String url;
    private final boolean loginRequired;
    private final Chaturbate chaturbate;
    private final int modelsPerPage = 90;

    public ChaturbateUpdateService(String url, boolean loginRequired, Chaturbate chaturbate) {
        this.url = url;
        this.loginRequired = loginRequired;
        this.chaturbate = chaturbate;
    }

    @Override
    protected Task<List<Model>> createTask() {
        return new Task<>() {
            @Override
            public List<Model> call() throws IOException {
                if (loginRequired && !chaturbate.credentialsAvailable()) {
                    return Collections.emptyList();
                } else {
                    int offset = (getPage() - 1) * modelsPerPage;
                    int limit = modelsPerPage;
                    String paginatedUrl = url + "&offset=" + offset + "&limit=" + limit;
                    LOG.debug("Fetching page {}", paginatedUrl);
                    if (loginRequired) {
                        SiteUiFactory.getUi(chaturbate).login();
                    }
                    Request request = new Request.Builder()
                            .url(paginatedUrl)
                            .header(ACCEPT, MIMETYPE_APPLICATION_JSON)
                            .header(ACCEPT_LANGUAGE, Locale.ENGLISH.getLanguage())
                            .header(X_REQUESTED_WITH, XML_HTTP_REQUEST)
                            .header(USER_AGENT, Config.getInstance().getSettings().httpUserAgent)
                            .build();
                    try (Response response = chaturbate.getHttpClient().execute(request)) {
                        if (response.isSuccessful()) {
                            return parseModels(response.body().string());
                        } else {
                            throw new HttpException(response.code(), response.message());
                        }
                    } // try
                } // if
            } // call
        };
    }

    private List<Model> parseModels(String body) {
        List<Model> models = new ArrayList<>();
        JSONObject json = new JSONObject(body);
        if (json.has("rooms")) {
            JSONArray jsonModels = json.getJSONArray("rooms");
            for (int i = 0; i < jsonModels.length(); i++) {
                var jsonModel = jsonModels.getJSONObject(i);
                try {
                    String name = jsonModel.getString("username");
                    Model model = chaturbate.createModel(name);
                    model.setDisplayName(name);
                    model.setPreview(jsonModel.optString("img"));
                    if (jsonModel.has("tags")) {
                        JSONArray tags = jsonModel.getJSONArray("tags");
                        for (int j = 0; j < tags.length(); j++) {
                            model.getTags().add(tags.optString(j));
                        }
                    }
                    if (jsonModel.has("subject")) {
                        String html = jsonModel.optString("subject");
                        model.setDescription(html2text(html));
                    }
                    models.add(model);
                } catch (Exception e) {
                    LOG.warn("Couldn't parse one of the models: {}", jsonModel, e);
                }
            }
            return models;
        } else {
            LOG.debug("Response was not successful: {}", json);
            return Collections.emptyList();
        }
    }

    public static String html2text(String html) {
        try {
            return Jsoup.parse(html).text();
        } catch (Exception ex) {
            return "";
        }
    }

    public void setUrl(String url) {
        this.url = url;
    }

}
