package ctbrec.ui.sites.winktv;

import static ctbrec.io.HttpConstants.*;

import ctbrec.Config;
import ctbrec.Model;
import ctbrec.io.HttpException;
import ctbrec.sites.winktv.WinkTv;
import ctbrec.sites.winktv.WinkTvModel;
import ctbrec.ui.SiteUiFactory;
import ctbrec.ui.tabs.PaginatedScheduledService;
import java.io.IOException;
import java.text.MessageFormat;
import java.time.Duration;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.Optional;
import java.util.function.Predicate;
import java.util.stream.Collectors;
import javafx.concurrent.Task;
import okhttp3.FormBody;
import okhttp3.Request;
import okhttp3.Response;
import org.json.JSONArray;
import org.json.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class WinkTvUpdateService extends PaginatedScheduledService {

    private static final Logger LOG = LoggerFactory.getLogger(WinkTvUpdateService.class);
    private static final String API_URL = "https://api.winktv.co.kr/v1/live";

    private WinkTv site;
    private String url;
    private static List<WinkTvModel> modelsList;
    private static Instant lastListInfoRequest = Instant.EPOCH;
    protected int modelsPerPage = 48;
    protected Predicate<WinkTvModel> filter;

    public WinkTvUpdateService(WinkTv site, Predicate<WinkTvModel> filter) {
        this.site = site;
        this.filter = filter;
    }

    @Override
    protected Task<List<Model>> createTask() {
        return new Task<List<Model>>() {
            @Override
            public List<Model> call() throws IOException {
                return getModelList().stream()
                        .filter(filter)
                        .skip((page - 1) * (long) modelsPerPage)
                        .limit(modelsPerPage)
                        .collect(Collectors.toList()); // NOSONAR
            }
        };
    }

    private List<WinkTvModel> getModelList() throws IOException {
        if (Duration.between(lastListInfoRequest, Instant.now()).getSeconds() < 30) {
            return Optional.ofNullable(modelsList).orElse(loadModelList());
        }
        modelsList = loadModelList();
        return Optional.ofNullable(modelsList).orElse(Collections.emptyList());
    }

    private List<WinkTvModel> loadModelList() throws IOException {
        LOG.debug("Fetching page {}", API_URL);
        lastListInfoRequest = Instant.now();
        FormBody body = new FormBody.Builder()
                .add("offset", "0")
                .add("limit", "500")
                .add("orderBy", "hot")
                .build();
        Request req = new Request.Builder()
                .url(API_URL)
                .header(USER_AGENT, Config.getInstance().getSettings().httpUserAgent)
                .header(ACCEPT, MIMETYPE_APPLICATION_JSON)
                .header(ACCEPT_LANGUAGE, Locale.ENGLISH.getLanguage())
                .header(X_REQUESTED_WITH, XML_HTTP_REQUEST)
                .header(REFERER, site.getBaseUrl() + "/")
                .header(ORIGIN, site.getBaseUrl())
                .post(body)
                .build();
        try (var response = site.getHttpClient().execute(req)) {
            if (response.isSuccessful()) {
                List<WinkTvModel> models = new ArrayList<>();
                var content = response.body().string();
                var json = new JSONObject(content);
                if (json.optBoolean("result")) {
                    var modelNodes = json.getJSONArray("list");
                    parseModels(modelNodes, models);
                }
                return models;
            } else {
                throw new HttpException(response.code(), response.message());
            }
        }
    }

    private void parseModels(JSONArray jsonModels, List<WinkTvModel> models) {
        for (var i = 0; i < jsonModels.length(); i++) {
            var m = jsonModels.getJSONObject(i);
            String name = m.optString("userId");
            WinkTvModel model = (WinkTvModel) site.createModel(name);
            model.setDisplayName(m.getString("userNick"));
            boolean isAdult = m.optBoolean("isAdult"); 
            model.setAdult(isAdult);
            if (isAdult && m.has("ivsThumbnail")) {
                model.setPreview(m.optString("ivsThumbnail"));
            } else {
                model.setPreview(m.optString("thumbUrl"));
            }
            boolean isLive = m.optBoolean("isLive");
            if (isLive) models.add(model);
        }
    }
    
    public void setFilter(Predicate<WinkTvModel> filter) {
        this.filter = filter;
    }
    
}
